import os
import shutil

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

import pytest
import keras
import skorch
import torch
import numpy as np

from keras import layers
from keras.datasets import cifar10
from torch import nn
from torch.nn import functional as F

from mia.estimators import AttackModelBundle
from mia.estimators import ShadowModelBundle
from mia.estimators import prepare_attack_data
from mia.serialization import BaseModelSerializer

torch.set_default_tensor_type("torch.FloatTensor")


WIDTH = 32
HEIGHT = 32
CHANNELS = 3
NUM_CLASSES = 10

SHADOW_DATASET_SIZE = 200


def to_categorical(y, num_classes):
    """ 1-hot encodes a tensor """
    return np.eye(num_classes, dtype="uint8")[y]


class CrossEntropyOneHot(object):
    def __init__(self):
        self.cel = nn.CrossEntropyLoss()

    def __call__(self, out, target):
        _, labels = target.max(dim=1)
        return self.cel(out, labels)


def keras_shadow_model_fn():
    model = keras.models.Sequential()

    model.add(
        layers.Conv2D(
            32,
            (3, 3),
            activation="relu",
            padding="same",
            input_shape=(WIDTH, HEIGHT, CHANNELS),
        )
    )
    model.add(layers.Conv2D(32, (3, 3), activation="relu"))
    model.add(layers.MaxPooling2D(pool_size=(2, 2)))
    model.add(layers.Dropout(0.25))

    model.add(layers.Conv2D(64, (3, 3), activation="relu", padding="same"))
    model.add(layers.Conv2D(64, (3, 3), activation="relu"))
    model.add(layers.MaxPooling2D(pool_size=(2, 2)))
    model.add(layers.Dropout(0.25))

    model.add(layers.Flatten())

    model.add(layers.Dense(512, activation="relu"))
    model.add(layers.Dropout(0.5))

    model.add(layers.Dense(NUM_CLASSES, activation="softmax"))
    model.compile("adam", loss="categorical_crossentropy", metrics=["accuracy"])

    return model


class ShadowNet(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(CHANNELS, 6, kernel_size=3)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = nn.Conv2d(6, 16, 5)
        self.fc1 = nn.Linear(16 * 5 * 5, 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, NUM_CLASSES)

    def forward(self, x, **kwargs):
        del kwargs  # Unused.

        x = x.view(-1, CHANNELS, WIDTH, HEIGHT)
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = x.view(-1, 16 * 5 * 5)
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


def torch_shadow_model_fn():
    model = skorch.NeuralNetClassifier(
        module=ShadowNet, max_epochs=5, criterion=CrossEntropyOneHot, train_split=None
    )
    return model


# TODO: Make mock tests for the serializer.
class SkorchSerializer(BaseModelSerializer):
    def save(self, model_id, model):
        print("Saved %s" % model_id)
        model.save_params(self.get_model_path(model_id))

    def load(self, model_id):
        model = self.model_fn()
        model.initialize()
        model.load_params(self.get_model_path(model_id))
        print("Loaded %s" % model_id)
        return model


@pytest.fixture
def torch_shadow_serializer():
    prefix = "__serialized_torch_shadow_test_dir"
    serializer = SkorchSerializer(torch_shadow_model_fn, prefix)
    yield serializer
    shutil.rmtree(prefix)


@pytest.fixture
def torch_attack_serializer():
    prefix = "__serialized_torch_attack_test_dir"
    serializer = SkorchSerializer(torch_attack_model_fn, prefix)
    yield serializer
    shutil.rmtree(prefix)


def keras_attack_model_fn():
    model = keras.models.Sequential()

    # Input layer
    model.add(layers.Dense(128, activation="relu", input_shape=(NUM_CLASSES,)))

    # Hidden layers
    model.add(layers.Dropout(0.3, noise_shape=None, seed=None))
    model.add(layers.Dense(64, activation="relu"))
    model.add(layers.Dropout(0.2, noise_shape=None, seed=None))
    model.add(layers.Dense(64, activation="relu"))

    # Output layer
    model.add(layers.Dense(1, activation="sigmoid"))
    model.compile("adam", loss="binary_crossentropy", metrics=["accuracy"])
    return model


class AttackNet(nn.Module):
    def __init__(self):
        super().__init__()
        self.fc1 = nn.Linear(NUM_CLASSES, 128)
        self.fc2 = nn.Linear(128, 64)
        self.fc3 = nn.Linear(64, 64)
        self.softmax = nn.Linear(64, 1)

    def forward(self, x, **kwargs):
        del kwargs  # Unused.

        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = F.relu(self.fc3(x))
        x = F.sigmoid(self.softmax(x))
        return x


def torch_attack_model_fn():
    model = skorch.NeuralNetClassifier(
        module=AttackNet, max_epochs=5, criterion=nn.BCELoss, train_split=None
    )
    return model


@pytest.mark.parametrize("num_models", [1, 3])
@pytest.mark.parametrize(
    "shadow_model_fn,model_serializer",
    [
        (keras_shadow_model_fn, None),
        (torch_shadow_model_fn, None),
        (torch_shadow_model_fn, pytest.lazy_fixture("torch_shadow_serializer")),
    ],
)
def test_shadow_models_are_created_and_data_is_transformed(
    data, shadow_model_fn, model_serializer, num_models
):

    (X_train, y_train), _ = data
    smb = ShadowModelBundle(
        shadow_model_fn,
        shadow_dataset_size=SHADOW_DATASET_SIZE,
        num_models=num_models,
        serializer=model_serializer,
    )
    assert not hasattr(smb, "shadow_models_")

    X_shadow, y_shadow = smb.fit_transform(
        X_train, y_train, fit_kwargs=dict(epochs=5, verbose=False)
    )

    # X_shadow are prediction vectors for random examples in X_train
    assert len(X_shadow) == len(y_shadow) == (2 * num_models * SHADOW_DATASET_SIZE)

    # X_shadow are concatenations of prediction vectors and true
    # class.
    assert X_shadow.shape[1] == 2 * NUM_CLASSES


@pytest.mark.parametrize("num_models", [3])
@pytest.mark.parametrize("max_models_for_transform", [1, 2])
@pytest.mark.parametrize(
    "shadow_model_fn,model_serializer",
    [
        (keras_shadow_model_fn, None),
        (torch_shadow_model_fn, None),
        (torch_shadow_model_fn, pytest.lazy_fixture("torch_shadow_serializer")),
    ],
)
def test_shadow_transform_with_custom_shadow_indices(
    data, shadow_model_fn, model_serializer, num_models, max_models_for_transform
):

    (X_train, y_train), _ = data
    smb = ShadowModelBundle(
        shadow_model_fn,
        shadow_dataset_size=SHADOW_DATASET_SIZE,
        num_models=num_models,
        serializer=model_serializer,
    )

    smb._fit(X_train, y_train, fit_kwargs=dict(epochs=5, verbose=False))

    shadow_indices = range(max_models_for_transform)
    X_shadow, y_shadow = smb._transform(shadow_indices=shadow_indices)
    assert X_shadow.shape[0] == 2 * max_models_for_transform * SHADOW_DATASET_SIZE
    assert y_shadow.shape[0] == 2 * max_models_for_transform * SHADOW_DATASET_SIZE


@pytest.mark.parametrize(
    "shadow_model_fn,shadow_model_serializer,"
    "attack_model_fn,attack_model_serializer",
    [
        (keras_shadow_model_fn, None, keras_attack_model_fn, None),
        (keras_shadow_model_fn, None, torch_attack_model_fn, None),
        (torch_shadow_model_fn, None, torch_attack_model_fn, None),
        (torch_shadow_model_fn, None, keras_attack_model_fn, None),
        (
            torch_shadow_model_fn,
            pytest.lazy_fixture("torch_shadow_serializer"),
            torch_attack_model_fn,
            None,
        ),
        (
            torch_shadow_model_fn,
            pytest.lazy_fixture("torch_shadow_serializer"),
            torch_attack_model_fn,
            pytest.lazy_fixture("torch_attack_serializer"),
        ),
        (
            torch_shadow_model_fn,
            pytest.lazy_fixture("torch_shadow_serializer"),
            keras_attack_model_fn,
            None,
        ),
    ],
)
def test_attack_models_are_created(
    data,
    shadow_model_fn,
    shadow_model_serializer,
    attack_model_fn,
    attack_model_serializer,
):

    (X_train, y_train), _ = data
    smb = ShadowModelBundle(
        shadow_model_fn,
        shadow_dataset_size=SHADOW_DATASET_SIZE,
        num_models=3,
        serializer=shadow_model_serializer,
    )
    X_shadow, y_shadow = smb.fit_transform(
        X_train, y_train, fit_kwargs=dict(epochs=5, verbose=False)
    )

    amb = AttackModelBundle(
        attack_model_fn, num_classes=NUM_CLASSES, serializer=attack_model_serializer
    )

    # Fit the attack models.
    amb.fit(X_shadow, y_shadow, fit_kwargs=dict(epochs=5, verbose=False))

    # Predict membership for some training data.
    membership_guesses = amb.predict(X_shadow[:100])
    assert membership_guesses.shape == (len(X_shadow[:100]),)


@pytest.mark.parametrize("shadow_model_fn", [torch_shadow_model_fn])
def test_prepare_attack_data(data, shadow_model_fn):
    (X_train, y_train), (X_test, y_test) = data
    clf = shadow_model_fn()
    clf.fit(X_train, y_train, epochs=3, verbose=False)
    X_attack, y_attack = prepare_attack_data(
        clf, (X_train[:100], y_train[:100]), (X_test[:100], y_test[:100])
    )
    assert X_attack.shape == (200, 20)
    assert y_attack.shape == (200,)
    assert np.all(y_attack[:100])
    assert not np.any(y_attack[100:])


def test_serializer_creates_dir():
    prefix = "__serializer_test_dir"
    serializer = BaseModelSerializer(model_fn=None, prefix=prefix)
    assert prefix in os.listdir(".")
    shutil.rmtree(prefix)



import copy

import torch
import time
import tqdm
import numpy as np
import os

from torch import nn
from torch.autograd import Variable
from torch.utils.data import TensorDataset, DataLoader

from .serialization import BaseModelSerializer


class ExpLrScheduler(object):
    """Decay learning rate by a factor every `lr_decay_every_epochs`.

    Based on https://discuss.pytorch.org/t/fine-tuning-squeezenet/3855/7
    """

    def __init__(
        self, init_lr=0.001, decay_factor=0.1, lr_decay_every_epochs=7, verbose=False
    ):
        self.init_lr = init_lr
        self.decay_factor = decay_factor
        self.lr_decay_every_epochs = lr_decay_every_epochs
        self.verbose = verbose

    def __call__(self, optimizer, epoch):
        lr = self.init_lr * (self.decay_factor ** (epoch // self.lr_decay_every_epochs))

        if self.verbose and (epoch % self.lr_decay_every_epochs == 0):
            print("LR is set to {}".format(lr))

        for param_group in optimizer.param_groups:
            param_group["lr"] = lr

        return optimizer


def _numpy_to_dataloader(X, y=None, *args, **kwargs):
    X = torch.from_numpy(X)
    if y is not None:
        y = torch.from_numpy(y)
    return _torch_to_dataloader(X, y, *args, **kwargs)


def _torch_to_dataloader(X, y=None, *args, **kwargs):
    tensors = [X]
    if y is not None:
        tensors.append(y)
    dataset = TensorDataset(*tensors)
    return DataLoader(dataset, *args, **kwargs)


def _input_to_dataloader(X, y=None, offset=None, max_examples=None, *args, **kwargs):
    if offset is None:
        offset = 0
    if max_examples is not None:
        max_examples = len(X)
    X_slice = X[offset:max_examples]
    y_slice = y[offset:max_examples] if y is not None else None
    if isinstance(X, np.ndarray):
        return _numpy_to_dataloader(X_slice, y_slice, *args, **kwargs)
    elif isinstance(X, torch.Tensor):
        return _torch_to_dataloader(X_slice, y_slice, *args, **kwargs)
    else:
        raise NotImplementedError()


class TorchWrapperSerializer(BaseModelSerializer):
    """Torch wrapper serializer."""

    def __init__(self, model_fn, prefix, verbose=False):
        super().__init__(model_fn, prefix)
        self.verbose = verbose

    def _get_val_data_path(self, model_id):
        return os.path.join(self._get_model_path(model_id, "__val_data"))

    def save(self, model_id, model):
        print("Saved %s" % model_id)
        torch.save(model.module_.state_dict(), self._get_model_path(model_id))
        torch.save(model.val_loader_, self._get_val_data_path(model_id))

    def load(self, model_id):
        model = self.model_fn()
        model.module_.load_state_dict(torch.load(self._get_model_path(model_id)))
        if self.verbose:
            print("Loaded %s" % model_id)
        return model


class TorchWrapper(object):
    """
    Simplified Keras/sklearn-like wrapper for a torch module.

    We know there's skorch, but it was a pain to debug.

    :param module: Torch module class
    :param criterion: Criterion class
    :param optimizer: Optimizer class
    :param dict module_params: Parameters to pass to the module
            on initialization.
    :param dict optimizer_params: Parameters to pass to the optimizer
            on initialization.
    :param lr_scheduler: Learning rate scheduler
    :param enable_cude: Whether to use CUDA
    :param ModelSerializer serializer: Model serializer to save the best model.
    """

    STATS_FMT = "[{:>5s}] loss: {:+.4f}, acc: {:.4f}"
    BEST_MODEL_ID = "__fit_best"

    def __init__(
        self,
        module,
        criterion,
        optimizer,
        module_params=None,
        optimizer_params=None,
        lr_scheduler=None,
        enable_cuda=True,
        serializer=None,
    ):
        self.module = module
        self.criterion = criterion
        self.optimizer = optimizer
        self.enable_cuda = enable_cuda
        self.serializer = serializer

        self.criterion_ = criterion()

        if module_params is None:
            module_params = {}
        self.module_ = module(**module_params)
        if enable_cuda:
            self.module_ = self.module_.cuda()

        if optimizer_params is None:
            optimizer_params = {}
        self.optimizer_ = optimizer(self.module_.parameters(), **optimizer_params)

        self.lr_scheduler = lr_scheduler
        if lr_scheduler is None:
            self.lr_scheduler = lambda opt, *args, **kwargs: opt

    def fit_step(self, batch, phase="train"):
        """
        Run a single training step.

        :param batch: A tuple of numpy batch examples and labels
        :param phase: Phase. One of ['train', 'val']. If in val, does not
                      update the model parameters.
        """
        self.module_.train(phase == "train")
        inputs, labels = batch
        batch_size = len(inputs)

        # Wrap them in Variable
        if self.enable_cuda:
            inputs, labels = Variable(inputs.cuda()), Variable(labels.cuda())
        else:
            inputs, labels = Variable(inputs), Variable(labels)

        # Zero the parameter gradients.
        self.optimizer_.zero_grad()

        # Forward pass.
        outputs = self.module_(inputs)
        _, preds = torch.max(outputs.data, 1)

        # Calculating the loss.
        loss = self.criterion_(outputs, labels)

        # Backward + optimize only if in training phase.
        if phase == "train":
            loss.backward()
            self.optimizer_.step()

        # Batch statistics.
        batch_loss = float(loss)
        num_correct_preds = float(torch.sum(preds == labels.data))

        return batch_loss, batch_size, num_correct_preds

    def fit(
        self,
        X,
        y=None,
        batch_size=32,
        epochs=20,
        shuffle=True,
        validation_split=None,
        validation_data=None,
        verbose=False,
    ):
        """
        Fit a torch classifier.

        :param X: Dataset
        :type X: ``numpy.ndarray`` or ``torch.Tensor``.
        :param y: Labels
        :param batch_size: Batch size
        :param epochs: Number of epochs to run the training
        :param shuffle: Whether to shuffle the dataset
        :param validation_split: Ratio of data to use for training. E.g., 0.7
        :param validation_data: If ``validation_split`` is not specified,
                the explicit validation dataset.
        :param verbose: Whether to output the progress report.

        TODO: Add custom metrics.
        """
        max_train_examples = None
        val_offset = None

        if validation_split is not None:
            max_train_examples = int((1 - validation_split) * len(X))
            val_offset = int(validation_split * len(X))
            validation_data = (X, y)

        train_loader = _input_to_dataloader(
            X,
            y,
            batch_size=batch_size,
            shuffle=shuffle,
            max_examples=max_train_examples,
        )

        phases = ["train"]
        if validation_data is not None:
            self.val_loader_ = val_loader = _input_to_dataloader(
                *validation_data,
                batch_size=batch_size,
                shuffle=shuffle,
                offset=val_offset
            )
            phases.append("val")
            best_val_loss = 0.0

        since = time.time()

        for epoch in range(epochs):
            if verbose:
                print("Epoch %d/%d" % (epoch + 1, epochs))

            # Each epoch has a training and validation phase.
            for phase in phases:
                if phase == "train":
                    self.optimizer_ = self.lr_scheduler(self.optimizer_, epoch)
                    data_iter = train_loader
                else:
                    data_iter = val_loader

                if verbose and phase == "train":
                    prog_bar = tqdm.tqdm(total=len(data_iter.dataset))

                epoch_dataset_size = 0
                running_loss = 0.0
                running_num_correct_preds = 0

                # Run through the data in minibatches.
                for data in data_iter:
                    batch_loss, batch_size, num_correct_preds = self.fit_step(
                        data, phase=phase
                    )

                    # Batch statistics.
                    epoch_dataset_size += batch_size
                    running_loss += batch_loss
                    running_num_correct_preds += num_correct_preds

                    if verbose and phase == "train":
                        batch_stats_str = TorchWrapper.STATS_FMT.format(
                            phase,
                            batch_loss / batch_size,
                            num_correct_preds / batch_size,
                        )
                        prog_bar.set_description(batch_stats_str)
                        prog_bar.update(batch_size)

                # Epoch statistics.
                epoch_loss = running_loss / epoch_dataset_size
                epoch_acc = running_num_correct_preds / epoch_dataset_size

                if verbose:
                    prog_bar.close()
                    epoch_stats_str = TorchWrapper.STATS_FMT.format(
                        phase, epoch_loss, epoch_acc
                    )
                    print(epoch_stats_str)

                # Determine if model is the best.
                if phase == "val" and self.serializer is not None:
                    if epoch_loss < best_loss:
                        best_loss = epoch_loss
                        # TODO: Needs more work. Check the temperature
                        #       scaling class. Neural Network should change.
                        self.serializer.save(TorchWrapper.BEST_MODEL_ID)
                        if verbose:
                            print("New best accuracy: %.4f" % best_loss)

        time_elapsed = time.time() - since
        print(
            "Training complete in {:.0f}m {:.0f}s".format(
                time_elapsed // 60, time_elapsed % 60
            )
        )

        # Load the best model.
        if "val" in phases and self.serializer is not None:
            model_wrapper = self.serializer.load(TorchWrapper.BEST_MODEL_ID)
            self.module_ = model_wrapper.module_
            self.val_loader_ = model_wrapper.val_loader_

        return self.module_

    def predict_proba(self, X, batch_size=32):
        """Get the confidence vector for an evaluation of a trained model.

        :param X: Data
        :param batch_size: Batch size

        TODO: Fix in case this is not one-hot.

        """
        data_iter = _input_to_dataloader(X, batch_size=batch_size)

        batch_outputs = []
        for batch in data_iter:
            x = batch[0]
            if self.enable_cuda:
                x = x.cuda()
            outputs = self.module_(x).data
            batch_outputs.append(outputs)
        return np.vstack(batch_outputs)

    def predict(self, X, batch_size=32):
        """Get the confidence vector for an evaluation of a trained model.

        :param X: Data
        :param batch_size: Batch size
        """
        return np.argmax(self.predict_proba(X, batch_size), axis=1)



import pytest
import keras

from keras.datasets import cifar10


@pytest.fixture(scope="module")
def data():
    (X_train, y_train), (X_test, y_test) = cifar10.load_data()
    y_train = keras.utils.to_categorical(y_train)
    y_test = keras.utils.to_categorical(y_test)
    X_train = X_train.astype("float32")
    X_test = X_test.astype("float32")
    y_train = y_train.astype("float32")
    y_test = y_test.astype("float32")
    X_train /= 255
    X_test /= 255
    return (X_train, y_train), (X_test, y_test)



"""
Scikit-like estimators for the attack model and shadow models.
"""

import sklearn
import numpy as np

from tqdm import tqdm


class ShadowModelBundle(sklearn.base.BaseEstimator):
    """
    A bundle of shadow models.

    :param model_fn: Function that builds a new shadow model
    :param shadow_dataset_size: Size of the training data for each shadow model
    :param num_models: Number of shadow models
    :param seed: Random seed
    :param ModelSerializer serializer: Serializer for the models. If None,
            the shadow models will be stored in memory. Otherwise, loaded
            and saved when needed.
    """

    MODEL_ID_FMT = "shadow_%d"

    def __init__(
        self, model_fn, shadow_dataset_size, num_models=20, seed=42, serializer=None
    ):
        super().__init__()
        self.model_fn = model_fn
        self.shadow_dataset_size = shadow_dataset_size
        self.num_models = num_models
        self.seed = seed
        self.serializer = serializer
        self._reset_random_state()

    def fit_transform(self, X, y, verbose=False, fit_kwargs=None):
        """Train the shadow models and get a dataset for training the attack.

        :param X: Data coming from the same distribution as the target
                  training data
        :param y: Data labels
        :param bool verbose: Whether to display the progressbar
        :param dict fit_kwargs: Arguments that will be passed to the fit call for
                each shadow model.

        .. note::
            Be careful when holding out some of the passed data for validation
            (e.g., if using Keras, passing `fit_kwargs=dict(validation_split=0.7)`).
            Such data will be marked as "used in training", whereas it was used for
            validation. Doing so may decrease the success of the attack.
        """
        self._fit(X, y, verbose=verbose, fit_kwargs=fit_kwargs)
        return self._transform(verbose=verbose)

    def _reset_random_state(self):
        self._prng = np.random.RandomState(self.seed)

    def _get_model_iterator(self, indices=None, verbose=False):
        if indices is None:
            indices = range(self.num_models)
        if verbose:
            indices = tqdm(indices)
        return indices

    def _get_model(self, model_index):
        if self.serializer is not None:
            model_id = ShadowModelBundle.MODEL_ID_FMT % model_index
            model = self.serializer.load(model_id)
        else:
            model = self.shadow_models_[model_index]
        return model

    def _fit(self, X, y, verbose=False, pseudo=False, fit_kwargs=None):
        """Train the shadow models.

        .. note::
        Be careful not to hold out some of the passed data for validation
        (e.g., if using Keras, passing `fit_kwargs=dict(validation_split=0.7)`).
        Such data will be incorrectly marked as "used in training", whereas
        it was not.

        :param X: Data coming from the same distribution as the target
                  training data
        :param y: Data labels
        :param bool verbose: Whether to display the progressbar
        :param bool pseudo: If True, does not fit the models
        :param dict fit_kwargs: Arguments that will be passed to the fit call for
                each shadow model.
        """
        self.shadow_train_indices_ = []
        self.shadow_test_indices_ = []

        if self.serializer is None:
            self.shadow_models_ = []

        fit_kwargs = fit_kwargs or {}
        indices = np.arange(X.shape[0])

        for i in self._get_model_iterator(verbose=verbose):
            # Pick indices for this shadow model.
            shadow_indices = self._prng.choice(
                indices, 2 * self.shadow_dataset_size, replace=False
            )
            train_indices = shadow_indices[: self.shadow_dataset_size]
            test_indices = shadow_indices[self.shadow_dataset_size :]
            X_train, y_train = X[train_indices], y[train_indices]
            self.shadow_train_indices_.append(train_indices)
            self.shadow_test_indices_.append(test_indices)

            if pseudo:
                continue

            # Train the shadow model.
            shadow_model = self.model_fn()
            shadow_model.fit(X_train, y_train, **fit_kwargs)
            if self.serializer is not None:
                self.serializer.save(ShadowModelBundle.MODEL_ID_FMT % i, shadow_model)
            else:
                self.shadow_models_.append(shadow_model)

        self.X_fit_ = X
        self.y_fit_ = y
        self._reset_random_state()
        return self

    def _pseudo_fit(self, X, y, verbose=False, fit_kwargs=None):
        self._fit(X, y, verbose=verbose, fit_kwargs=fit_kwargs, pseudo=True)

    def _transform(self, shadow_indices=None, verbose=False):
        """Produce in/out data for training the attack model.

        :param shadow_indices: Indices of the shadow models to use
                for generating output data.
        :param verbose: Whether to show progress
        """
        shadow_data_array = []
        shadow_label_array = []

        model_index_iter = self._get_model_iterator(
            indices=shadow_indices, verbose=verbose
        )

        for i in model_index_iter:
            shadow_model = self._get_model(i)
            train_indices = self.shadow_train_indices_[i]
            test_indices = self.shadow_test_indices_[i]

            train_data = self.X_fit_[train_indices], self.y_fit_[train_indices]
            test_data = self.X_fit_[test_indices], self.y_fit_[test_indices]
            shadow_data, shadow_labels = prepare_attack_data(
                shadow_model, train_data, test_data
            )

            shadow_data_array.append(shadow_data)
            shadow_label_array.append(shadow_labels)

        X_transformed = np.vstack(shadow_data_array).astype("float32")
        y_transformed = np.hstack(shadow_label_array).astype("float32")
        return X_transformed, y_transformed


def prepare_attack_data(model, data_in, data_out):
    """
    Prepare the data in the attack model format.

    :param model: Classifier
    :param (X, y) data_in: Data used for training
    :param (X, y) data_out: Data not used for training

    :returns: (X, y) for the attack classifier
    """
    X_in, y_in = data_in
    X_out, y_out = data_out
    y_hat_in = model.predict_proba(X_in)
    y_hat_out = model.predict_proba(X_out)

    labels = np.ones(y_in.shape[0])
    labels = np.hstack([labels, np.zeros(y_out.shape[0])])
    # TODO: this does not work for non-one-hot labels.
    # data = np.hstack([y_hat_in, y_in])
    data = np.c_[y_hat_in, y_in]
    data = np.vstack([data, np.c_[y_hat_out, y_out]])
    return data, labels


class AttackModelBundle(sklearn.base.BaseEstimator):
    """
    A bundle of attack models, one for each target model class.

    :param model_fn: Function that builds a new shadow model
    :param num_classes: Number of classes
    :param ModelSerializer serializer: Serializer for the models. If not None,
            the models will not be stored in memory, but rather loaded
            and saved when needed.
    :param class_one_hot_encoded: Whether the shadow data uses one-hot encoded
            class labels.
    """

    MODEL_ID_FMT = "attack_%d"

    def __init__(
        self, model_fn, num_classes, serializer=None, class_one_hot_coded=True
    ):
        self.model_fn = model_fn
        self.num_classes = num_classes
        self.serializer = serializer
        self.class_one_hot_coded = class_one_hot_coded

    def fit(self, X, y, verbose=False, fit_kwargs=None):
        """Train the attack models.

        :param X: Shadow predictions coming from
                  :py:func:`ShadowBundle.fit_transform`.
        :param y: Ditto
        :param verbose: Whether to display the progressbar
        :param fit_kwargs: Arguments that will be passed to the fit call for
                each attack model.
        """
        X_total = X[:, : self.num_classes]
        classes = X[:, self.num_classes :]

        datasets_by_class = []
        data_indices = np.arange(X_total.shape[0])
        for i in range(self.num_classes):
            if self.class_one_hot_coded:
                class_indices = data_indices[np.argmax(classes, axis=1) == i]
            else:
                class_indices = data_indices[np.squeeze(classes) == i]

            datasets_by_class.append((X_total[class_indices], y[class_indices]))

        if self.serializer is None:
            self.attack_models_ = []

        dataset_iter = datasets_by_class
        if verbose:
            dataset_iter = tqdm(dataset_iter)
        for i, (X_train, y_train) in enumerate(dataset_iter):
            model = self.model_fn()
            fit_kwargs = fit_kwargs or {}
            model.fit(X_train, y_train, **fit_kwargs)

            if self.serializer is not None:
                model_id = AttackModelBundle.MODEL_ID_FMT % i
                self.serializer.save(model_id, model)
            else:
                self.attack_models_.append(model)

    def _get_model(self, model_index):
        if self.serializer is not None:
            model_id = AttackModelBundle.MODEL_ID_FMT % model_index
            model = self.serializer.load(model_id)
        else:
            model = self.attack_models_[model_index]
        return model

    def predict_proba(self, X):
        result = np.zeros((X.shape[0], 2))
        shadow_preds = X[:, : self.num_classes]
        classes = X[:, self.num_classes :]

        data_indices = np.arange(shadow_preds.shape[0])
        for i in range(self.num_classes):
            model = self._get_model(i)
            if self.class_one_hot_coded:
                class_indices = data_indices[np.argmax(classes, axis=1) == i]
            else:
                class_indices = data_indices[np.squeeze(classes) == i]

            membership_preds = model.predict(shadow_preds[class_indices])
            for j, example_index in enumerate(class_indices):
                prob = np.squeeze(membership_preds[j])
                result[example_index, 1] = prob
                result[example_index, 0] = 1 - prob

        return result

    def predict(self, X):
        probs = self.predict_proba(X)[:, 1]
        return probs > 0.5



__version__ = "0.1.2"
__title__ = "mia"
__author__ = "Bogdan Kulynych, Mohammad Yaghini"
__email__ = "hello@bogdankulynych.me, mohammad.yaghini@epfl.ch"
__copyright__ = "2018 by Bogdan Kulynych (EPFL SPRING Lab)"
__description__ = "A library for running membership inference attacks against ML models"
__url__ = "https://github.com/spring-epfl/mia"
__license__ = "MIT"



#!/usr/bin/env python
import os
import re

from setuptools import setup


INSTALL_REQUIRES = ["numpy", "scipy", "scikit-learn", "torch", "tqdm"]

SETUP_REQUIRES = ["pytest-runner"]

TEST_REQUIRES = ["pytest", "pytest-lazy-fixture", "tensorflow", "skorch", "keras"]

DEV_REQUIRES = TEST_REQUIRES + ["sphinx", "sphinx_rtd_theme", "black"]


here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, "README.rst")) as f:
    long_description = f.read()


with open(os.path.join(here, "mia/__init__.py")) as f:
    matches = re.findall(r"(__.+__) = \"(.*)\"", f.read())
    for var_name, var_value in matches:
        globals()[var_name] = var_value


setup(
    name=__title__,
    version=__version__,
    description=__description__,
    long_description=long_description,
    author=__author__,
    author_email=__email__,
    packages=["mia"],
    license=__license__,
    url=__url__,
    install_requires=INSTALL_REQUIRES,
    setup_requires=SETUP_REQUIRES,
    tests_require=TEST_REQUIRES,
    extras_require={"dev": DEV_REQUIRES, "test": TEST_REQUIRES},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: MIT License",
    ],
)



import os
import abc


class BaseModelSerializer(object):
    """
    ABC class for a model serializer.

    :param model_fn: Function that builds a new model
    :param prefix: Path to the directory where models will be saved.
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self, model_fn, prefix=".", *args, **kwargs):
        self.model_fn = model_fn
        self.models_path = prefix
        os.makedirs(self.models_path, exist_ok=True)

    def get_model_path(self, model_id):
        """Get the path to the model with given ID."""
        return os.path.join(self.models_path, model_id)

    @abc.abstractmethod
    def load(self, model_id):
        pass

    @abc.abstractmethod
    def save(self, model, model_id):
        pass



import pytest
import torch
import numpy as np

from torch import nn
from torch.nn import functional as F

from mia.wrappers import TorchWrapper, ExpLrScheduler

torch.set_default_tensor_type("torch.FloatTensor")

WIDTH = 32
HEIGHT = 32
CHANNELS = 3
NUM_CLASSES = 10


class Net(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(CHANNELS, 6, kernel_size=3)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = nn.Conv2d(6, 16, 5)
        self.fc1 = nn.Linear(16 * 5 * 5, 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, NUM_CLASSES)

    def forward(self, x, **kwargs):
        del kwargs  # Unused.

        x = x.view(-1, CHANNELS, WIDTH, HEIGHT)
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = x.view(-1, 16 * 5 * 5)
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


@pytest.mark.parametrize(
    "lr_scheduler", [None, ExpLrScheduler(lr_decay_every_epochs=1)]
)
@pytest.mark.parametrize("enable_cuda", [False])
@pytest.mark.parametrize("use_torch_arrays", [True, False])
def test_torch_wrapper_fit(data, lr_scheduler, enable_cuda, use_torch_arrays):
    model = TorchWrapper(
        Net, nn.CrossEntropyLoss, torch.optim.Adam, enable_cuda=enable_cuda
    )

    (X_train, y_train), (X_test, y_test) = data
    y_train = np.argmax(y_train, axis=1)
    y_test = np.argmax(y_test, axis=1)

    if use_torch_arrays:
        X_train = torch.tensor(X_train)
        y_train = torch.tensor(y_train)

    model.fit(X_train, y_train, epochs=2, verbose=True, validation_split=0.1)

    # Expected accuracy is greater than 30%.
    assert np.mean(model.predict(X_test) == y_test) > 0.3
